package aoc2023

import println
import readInput

/*
--- Day 18: Lavaduct Lagoon ---
Thanks to your efforts, the machine parts factory is one of the first factories up and running since the lavafall came back. However, to catch up with the large backlog of parts requests, the factory will also need a large supply of lava for a while; the Elves have already started creating a large lagoon nearby for this purpose.
However, they aren't sure the lagoon will be big enough; they've asked you to take a look at the dig plan (your puzzle input). For example:
R 6 (#70c710)
D 5 (#0dc571)
L 2 (#5713f0)
D 2 (#d2c081)
R 2 (#59c680)
D 2 (#411b91)
L 5 (#8ceee2)
U 2 (#caa173)
L 1 (#1b58a2)
U 2 (#caa171)
R 2 (#7807d2)
U 3 (#a77fa3)
L 2 (#015232)
U 2 (#7a21e3)

The digger starts in a 1 meter cube hole in the ground. They then dig the specified number of meters up (U), down (D), left (L), or right (R), clearing full 1 meter cubes as they go. The directions are given as seen from above, so if "up" were north, then "right" would be east, and so on. Each trench is also listed with the color that the edge of the trench should be painted as an RGB hexadecimal color code.
When viewed from above, the above example dig plan would result in the following loop of trench (#) having been dug out from otherwise ground-level terrain (.):
#######
#.....#
###...#
..#...#
..#...#
###.###
#...#..
##..###
.#....#
.######

At this point, the trench could contain 38 cubic meters of lava. However, this is just the edge of the lagoon; the next step is to dig out the interior so that it is one meter deep as well:
#######
#######
#######
..#####
..#####
#######
#####..
#######
.######
.######

Now, the lagoon can contain a much more respectable 62 cubic meters of lava. While the interior is dug out, the edges are also painted according to the color codes in the dig plan.
The Elves are concerned the lagoon won't be large enough; if they follow their dig plan, how many cubic meters of lava could it hold?

*/
fun main() {

    val inlineTestInput = """
R 6 (#70c710)
D 5 (#0dc571)
L 2 (#5713f0)
D 2 (#d2c081)
R 2 (#59c680)
D 2 (#411b91)
L 5 (#8ceee2)
U 2 (#caa173)
L 1 (#1b58a2)
U 2 (#caa171)
R 2 (#7807d2)
U 3 (#a77fa3)
L 2 (#015232)
U 2 (#7a21e3)
"""

    data class DigInst(val cmd: String, val len: Int)

    fun calcArea(insts: List<DigInst>): Long {
        val polygon = ArrayList<Pair<Long, Long>>()
        var x = 0L
        var y = 0L
        var sumOutline = 0L
        for (i in insts) {
            when (i.cmd) {
                "R" -> x += i.len
                "D" -> y += i.len
                "L" -> x -= i.len
                "U" -> y -= i.len
            }
            sumOutline += i.len
            polygon.add(x to y)
        }
        var area = 0L
        for (i in 0 until polygon.size) {
            val j = (i + 1) % polygon.size
            area += (polygon[i].second + polygon[j].second) * (polygon[i].first - polygon[j].first)
        }
        return (area + sumOutline) / 2 + 1
    }

    fun part1(input: List<String>): Long {
        val insts = input.map {
            val (cmd, len, color) = "([RDLU]) (\\d+) \\(#([a-z0-f]+)\\)".toRegex().matchEntire(it)!!.destructured
            DigInst(cmd, len.toInt())
        }
        return calcArea(insts)
    }

    fun part2(input: List<String>): Long {
        val insts = input.map {
            val (cmd, len, color) = "([RDLU]) (\\d+) \\(#([a-z0-f]+)\\)".toRegex().matchEntire(it)!!.destructured
            DigInst("RDLU"[(color.toInt(16) and 3)].toString(), color.toInt(16) shr 4)
        }
        return calcArea(insts)
    }

    // test if implementation meets criteria from the description, like:
    val testInput = inlineTestInput.trim().reader().readLines()
    //val testInput = readInput("aoc2023/Day18_test")
    val testInputPart1Result = part1(testInput)
    println("Part 1 Test: $testInputPart1Result")
    val testInputPart2Result = part2(testInput)
    println("Part 2 Test: $testInputPart2Result")
    check(testInputPart1Result == 62L)
    check(testInputPart2Result == 952408144115L)

    val input = readInput("aoc2023/Day18")
    part1(input).println()
    part2(input).println()
}
